from __future__ import annotations

from charz_core import Scene

from ._grouping import Group
from ._annotations import AnimatedNode


# Define additional frame tasks for `Scene`


def progress_animations(current_scene: Scene) -> None:
    """Update animations for all animated nodes in the current scene."""
    for animated_node in current_scene.get_group_members(
        Group.ANIMATED,
        type_hint=AnimatedNode,
    ):
        animated_node.progress_animation()


# Register additional frame tasks for `Scene`
Scene.frame_tasks[70] = progress_animations



"""
Custom Annotations for `charz`
------------------------------

This file contains private annotations used across this package.

Whenever there is a "?" comment,
it means a type may or may not implement that field or mixin class.
"""

from __future__ import annotations as _annotations

from typing import (
    TypeVar as _TypeVar,
    TypeAlias as _TypeAlias,
    Protocol as _Protocol,
    runtime_checkable as _runtime_checkable,
    TYPE_CHECKING as _TYPE_CHECKING,
)

from charz_core import (
    Vec2 as _Vec2,
    Vec2i as _Vec2i,
)
from charz_core._annotations import (
    Node,
    TransformComponent,
    TransformNode,
    Engine as _CoreEngine,
)
from colex import ColorValue as _ColorValue

if _TYPE_CHECKING:
    from charz_core import Self as _Self
    from ._clock import Clock as _Clock
    from ._screen import Screen as _Screen
    from ._components._collision import Hitbox as _Hitbox
    from ._animation import (
        Animation as _Animation,
        AnimationSet as _AnimationSet,
    )

T = _TypeVar("T")
Number = _TypeVar("Number", float, int)
_T_contra = _TypeVar("_T_contra", contravariant=True)
Char: _TypeAlias = str  # Signifying `str` of length `1`


@_runtime_checkable
class FileLike(_Protocol[_T_contra]):
    def write(self, stream: _T_contra, /) -> object: ...
    def flush(self, /) -> None: ...
    def fileno(self, /) -> int: ...


class Engine(_CoreEngine, _Protocol):
    clock: _Clock
    screen: _Screen


class TextureComponent(_Protocol):
    texture: list[str]
    unique_texture: bool
    visible: bool
    centered: bool
    z_index: int
    transparency: Char | None

    def with_texture(self, texture_or_line: list[str] | str | Char, /) -> _Self: ...
    def with_unique_texture(self) -> _Self: ...
    def with_visibility(self, state: bool = True, /) -> _Self: ...
    def with_centering(self, state: bool = True, /) -> _Self: ...
    def with_z_index(self, z_index: int, /) -> _Self: ...
    def with_transparency(self, char: Char | None, /) -> _Self: ...
    def hide(self) -> None: ...
    def show(self) -> None: ...
    def is_globally_visible(self) -> bool: ...
    def get_texture_size(self) -> _Vec2i: ...


class TextureNode(
    TextureComponent,
    TransformNode,
    _Protocol,
): ...


class ColorComponent(_Protocol):
    color: _ColorValue | None

    def with_color(
        self,
        color: _ColorValue | None,
        /,
    ) -> _Self: ...


class ColorNode(
    ColorComponent,
    TextureNode,
    _Protocol,
): ...


class Renderable(
    # `ColorComponent`?
    TextureNode,
    _Protocol,
): ...


class AnimatedComponent(_Protocol):
    animations: _AnimationSet
    current_animation: _Animation | None
    repeat: bool
    is_playing: bool

    def with_animations(self, **animations: _Animation) -> _Self: ...
    def with_animation(
        self,
        animation_name: str,
        animation: _Animation,
        /,
    ) -> _Self: ...
    def with_repeat(self, state: bool = True, /) -> _Self: ...
    def add_animation(
        self,
        animation_name: str,
        animation: _Animation,
    ) -> None: ...
    def play(self, animation_name: str, /) -> None: ...
    def play_backwards(self, animation_name: str, /) -> None: ...
    def progress_animation(self) -> None: ...


class AnimatedNode(
    # `ColorComponent`?
    AnimatedComponent,
    TextureNode,
    _Protocol,
): ...


class ColliderComponent(_Protocol):
    hitbox: _Hitbox

    def with_hitbox(self, hitbox: _Hitbox, /) -> _Self: ...
    def get_colliders(self) -> list[ColliderNode]: ...
    def get_corner_points(self) -> tuple[_Vec2, _Vec2, _Vec2, _Vec2]: ...
    def is_colliding_with(self, colldier_node: ColliderNode, /) -> bool: ...
    def is_colliding(self) -> bool: ...


class ColliderNode(
    # `ColorComponent`?
    # `TextureComponent`?
    ColliderComponent,
    TransformNode,
    _Protocol,
): ...



from __future__ import annotations
from typing_extensions import Self

__all__ = ("SimpleMovementComponent",)

from typing import TYPE_CHECKING, NoReturn, Any

from charz_core import Scene, TransformComponent, Vec2, group

from .._time import Time
from .._grouping import Group

if TYPE_CHECKING:
    import keyboard
else:
    keyboard = None

loaded_simple_movement = False


# Use lazy loading to allow for not using `keyboard` module, as long as it is never used
def __getattr__(name: str) -> type[SimpleMovementComponent] | NoReturn:
    global keyboard  # noqa: PLW0603
    if keyboard is None:
        try:
            import keyboard as _keyboard  # noqa: PLC0415

            keyboard = _keyboard
        except ModuleNotFoundError as error:
            raise ModuleNotFoundError(
                "Module 'keyboard' was not found,"
                " use 'charz' with 'keyboard' or 'all' feature flag,"
                " like depending on 'charz[keyboard]' in 'pyproject.toml'"
            ) from error

    # Add frame task - Once the component is loaded
    global loaded_simple_movement  # noqa: PLW0603
    if not loaded_simple_movement:
        loaded_simple_movement = True
        Scene.frame_tasks[65] = update_moving_nodes

    # FIXME: Argument `name` has sometimes a value of `__path__`
    if name not in __all__:
        raise AttributeError(f"Module '{__name__}' has no attribute '{name}'")

    return SimpleMovementComponent


@group(Group.MOVEMENT)
class SimpleMovementComponent:  # Component (mixin class)
    """`SimpleMovementComponent` mixin class for node.

    It provides basic movement functionality for a node,
    and allows the node to move in 2D space using the `WASD` keys.

    Example:

    Player class with the ability to move using `WASD`:

    ```python
    from charz import Sprite
    from charz import SimpleMovementComponent

    class Player(Sprite, SimpleMovementComponent):
        texture = ["@"]
    ```

    Attributes:
        `speed`: `float` - The speed of the node's movement per second (`units/s`).
            Defaults to `16` `units/s`, which is a reasonable speed for most games.
            You can change this value to make the node move faster or slower,
            by overriding this attribute in your node class as a class attribute.
            When `use_delta_time` is `False`, the unit will be `units/frame`.
            Movement direction is normalized when `normalize_movement` is `True`.
        `use_delta_time`: `bool` - Whether to use delta time for movement.
            Defaults to `True`; movement will be frame-rate independent.
        `normalize_movement`: `bool` - Whether to normalize the movement direction vector.
            Defaults to `True`; movement direction will always have a length of `1`.

    Methods:
        `is_moving_left`
        `is_moving_right`
        `is_moving_up`
        `is_moving_down`
        `get_movement_direction`
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        # Workaround when using `@group`
        return super().__new__(cls, *args, **kwargs)

    speed: float = 16
    use_delta_time: bool = True
    normalize_movement: bool = True

    def is_moving_left(self) -> bool:
        """Check if the node is moving left.

        Override implementation to change the key used for moving left.

        Returns:
            `bool`: `True` if the node is moving left, `False` otherwise.
        """
        return keyboard.is_pressed("a")

    def is_moving_right(self) -> bool:
        """Check if the node is moving right.

        Override implementation to change the key used for moving right.

        Returns:
            `bool`: `True` if the node is moving right, `False` otherwise.
        """
        return keyboard.is_pressed("d")

    def is_moving_up(self) -> bool:
        """Check if the node is moving up.

        Override implementation to change the key used for moving up.

        Returns:
            `bool`: `True` if the node is moving up, `False` otherwise.
        """
        return keyboard.is_pressed("w")

    def is_moving_down(self) -> bool:
        """Check if the node is moving down.

        Override implementation to change the key used for moving down.

        Returns:
            `bool`: `True` if the node is moving down, `False` otherwise.
        """
        return keyboard.is_pressed("s")

    def get_movement_direction(self) -> Vec2:
        """Get the movement direction of the node.

        This method returns a `Vec2` object representing the direction
        of movement based on the current input.

        `NOTE` The returned vector is **not** normalized,
        meaning it can have a length greater than `1`.

        Returns:
            `Vec2`: Raw direction vector.
        """
        return Vec2(
            self.is_moving_right() - self.is_moving_left(),
            self.is_moving_down() - self.is_moving_up(),
        )

    def update_movement(self) -> None:
        """Custom update method for the node.

        Automatically handles checking for movement input,
        and moving the node accordingly.
        """
        assert isinstance(self, TransformComponent), (
            f"Node {self} missing `TransformComponent`"
        )
        direction = self.get_movement_direction()
        if self.normalize_movement:
            direction = direction.normalized()

        distance_delta = direction * self.speed
        if self.use_delta_time:
            distance_delta *= Time.delta

        self.position += distance_delta


# Define lazy added scene frame task


def update_moving_nodes(current_scene: Scene) -> None:
    """Update moving nodes in the current scene."""
    for moving_node in current_scene.get_group_members(
        Group.MOVEMENT,
        type_hint=SimpleMovementComponent,
    ):
        moving_node.update_movement()



from __future__ import annotations

from colex import ColorValue
from charz_core import Self


class ColorComponent:  # Component (mixin class)
    r"""`ColorComponent` mixin class for node.

    Example:

    Setting color of node. Since `Sprite` is composed using
    `ColorComponent`, `TextureComponent` and `Node2D`,
    the easiest will be to use a `Sprite` as a base:

    ```python
    import colex  # Color library used
    from charz import Sprite

    class PurpleMonster(Sprite):
        color = colex.PURPLE
        texture = [
            "   %",
            "~' | '~",
            "  / \\",
        ]
    ```

    Attributes:
        `color`: `ColorValue | None` - Optional color value for the node.
    """

    color: ColorValue | None = None

    def with_color(self, color: ColorValue | None, /) -> Self:
        self.color = color
        return self



from __future__ import annotations

from math import cos, sin
from dataclasses import dataclass
from copy import deepcopy
from typing import Any

from charz_core import Scene, TransformComponent, Vec2, Self, group

from .._grouping import Group
from .._annotations import ColliderNode


@dataclass(kw_only=True)
class Hitbox:
    """Hitbox dataclass for collision shape data.

    Example:

    Creating and centering `Hitbox` to a new collision node:

    ```python
    # collision_box.py
    from charz import Sprite, CollisionComponent, Hitbox, Vec2

    class CollisionBox(Sprite, CollisionComponent):
        hitbox = Hitbox(
            size=Vec2(5, 3),
            centered=True,  # Centering of collision hitbox
        )
        centered = True  # Centering of texture
        texture = [
            "#####",
            "#####",
            "#####",
        ]
    ```

    Attributes:
        `size`: `Vec2` - Width and height of the hitbox.
        `centered`: `bool` - Whether hitbox is centered around the node's global position.
            Defaults to `False`, meaning the hitbox starts at the node's position,
            and expanding to the right and downwards.
        `disabled`: `bool` - Whether collision with node is disabled.
            Defaults to `False`, meaning collision is active on with node.
        `margin`: `float` - Inverse margin around the hitbox for collision detection.
            Defaults to `1`, and should not be smaller than `1e-2`.
    """

    size: Vec2
    centered: bool = False
    disabled: bool = False
    margin: float = 1.0


@group(Group.COLLIDER)
class ColliderComponent:  # Component (mixin class)
    """`ColliderComponent` mixin class for node.

    Assign this component to a node to enable collision detection.
    All other collider components will then do collision detection against this node,
    when `is_colliding` and `get_colliders` is called.

    You can also use `is_colliding_with` for more fine-grained control.
    *Custom collision checks* can therefore be implemented by **overriding
    this method in a subclass**.

    Examples:

    Creating a boxes with collision, then printing the ones that collide:

    ```python
    # collision_box.py
    import colex
    from charz import Sprite, CollisionComponent, Hitbox, Vec2

    class CollisionBox(Sprite, CollisionComponent):
        hitbox = Hitbox(size=Vec2(5, 3))  # Usually matches the size of `texture`
        texture = [
            "#####",
            "#####",
            "#####",
        ]

    # main.py
    from charz import Engine
    from .collision_box import CollisionBox

    class MyGame(Engine):
        def __init__(self) -> None:
            self.box1 = CollisionBox(position=Vec2(2, 5), color=colex.RED)
            self.box2 = CollisionBox(position=Vec2(4, 7), color=colex.BLUE)
            self.box3 = CollisionBox(position=Vec2(5, 9), color=colex.GREEN)
            print(self.box2.get_colliders())

    # Prints out
    >>> ['CollisionBox(#0:Vec2(2, 5):0R:5x3:None)',
         'CollisionBox(#2:Vec2(5, 9):0R:5x3:None)']
    ```

    Filtering collision results, and deleting if collision occurs:

    ```python
    # Extending the last example...

    class Lethal: ...  # This works as a "tag" that can be detected using `isinstance`

    class LethalBox(Lethal, CollisionBox): ...

    box = CollisionBox(position=Vec2(4, 7))
    LethalBox(position=Vec2(3, 6))  # Remember: Reference not needed to create node

    for collider in box.get_colliders():
        if isinstance(collider, Lethal):
            print("Killed by", collider)
            box.queue_free()

    # Prints out
    >>> 'Killed by LethalBox(#1:Vec2(3, 6):0R:5x3:None)'
    ```

    Attributes:
        `hitbox`: `Hitbox` - The hitbox data for collision detection.
        `disabled`: `bool` - Whether the collider is disabled.

    Methods:
        `get_colliders`
        `is_colliding`
        `is_colliding_with`
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        if (class_hitbox := getattr(instance, "hitbox", None)) is not None:
            instance.hitbox = deepcopy(class_hitbox)
        else:
            instance.hitbox = Hitbox(size=Vec2.ZERO)
        return instance

    hitbox: Hitbox

    def with_hitbox(self, hitbox: Hitbox, /) -> Self:
        """Chained method to set the hitbox.

        Args:
            hitbox (Hitbox): The hitbox to set.

        Returns:
            Self: Same node instance.
        """
        self.hitbox = hitbox
        return self

    def get_colliders(self) -> list[ColliderNode]:
        """Get a list of colliders that this node is colliding with.

        This method iterates through all nodes in the `Group.Collider` group and checks
        if this node is colliding with any of them.

        Returns:
            list[ColliderNode]: List of colliders that this node is colliding with.
        """
        assert isinstance(self, ColliderComponent)
        nodes_collided_with: list[ColliderNode] = []
        # NOTE: Iterate `dict_values` instead of creating a `list` for speed
        for node in Scene.current.groups[Group.COLLIDER].values():
            if self is node:
                continue
            # Ignoring incorrect type because group `Group.Collider`
            # should only contain `ColliderNode` instances
            if self.is_colliding_with(node):  # type: ignore
                nodes_collided_with.append(node)  # type: ignore
        return nodes_collided_with

    def is_colliding(self) -> bool:
        """Check if this node is colliding with any other collider node.

        This method iterates through all nodes in the `Group.Collider` group and checks
        if this node is colliding with any of them.

        Returns:
            bool: Whether this node is colliding with any other collider node.
        """
        assert isinstance(self, ColliderComponent)
        for node in Scene.current.groups[Group.COLLIDER].values():
            if self is node:
                continue
            # Ignoring incorrect type because group `Group.Collider`
            # should only contain `ColliderNode` instances
            if self.is_colliding_with(node):  # type: ignore
                return True
        return False

    def is_colliding_with(self, collider_node: ColliderNode, /) -> bool:
        """Check if this node is colliding with another collider node.

        Uses SAT (Separating Axis Theorem).

        `NOTE` Does not yet fully support rotated hitboxes.

        Args:
            collider_node (ColliderNode): The other collider node to check collision with.

        Returns:
            bool: Whether this node is colliding with the other collider node.
        """
        if self.hitbox.disabled or collider_node.hitbox.disabled:
            return False

        corners_a = self.get_corner_points()
        corners_b = collider_node.get_corner_points()

        # Axes to test: x and y
        axes = [Vec2(1, 0), Vec2(0, 1)]

        for axis in axes:
            min_a, max_a = self._get_projection_range(corners_a, axis)
            min_b, max_b = self._get_projection_range(corners_b, axis)
            # Hitbox margin is negative space inside the hitbox,
            # extending from the edges
            if (
                max_a - self.hitbox.margin < min_b
                or max_b - collider_node.hitbox.margin < min_a
            ):
                return False  # Separating axis found

        return True  # No separating axis found, collision detected

    def get_corner_points(self) -> tuple[Vec2, Vec2, Vec2, Vec2]:
        assert isinstance(self, TransformComponent), (
            f"Node {self} missing `TransformComponent`"
        )
        global_position = self.global_position
        global_rotation = self.global_rotation
        hitbox_size = self.hitbox.size

        # Center the hitbox if needed
        if self.hitbox.centered:
            global_position = global_position - hitbox_size / 2

        # Define corners relative to position
        corners = (
            Vec2.ZERO,
            Vec2(hitbox_size.x, 0),
            hitbox_size,
            Vec2(0, hitbox_size.y),
        )

        # Rotate corners around the hitbox center
        if global_rotation != 0.0:
            center = global_position + hitbox_size / 2
            rotated = []
            for corner in corners:
                relative = global_position + corner - center
                rotated_corner = (
                    Vec2(
                        relative.x * cos(global_rotation)
                        - relative.y * sin(global_rotation),
                        relative.x * sin(global_rotation)
                        + relative.y * cos(global_rotation),
                    )
                    + center
                )
                rotated.append(rotated_corner)
            return tuple(rotated)
        else:
            return (  # Expanded to pass type checking
                global_position + corners[0],
                global_position + corners[1],
                global_position + corners[2],
                global_position + corners[3],
            )

    @staticmethod
    def _get_projection_range(
        corners: tuple[Vec2, Vec2, Vec2, Vec2],
        axis: Vec2,
    ) -> tuple[float, float]:
        projections = [corner.dot(axis) for corner in corners]
        return (min(projections), max(projections))



"""
Charz
=====

An object oriented terminal game engine

Includes
--------

- Annotations
  - `ColorValue`  (from package `colex`)
  - `Self`        (from standard `typing` or from package `typing-extensions`)
- Math (from package `linflex`)
  - `lerp`
  - `sign`
  - `clamp`
  - `move_toward`
  - `Vec2`
  - `Vec2i`
  - `Vec3`
- Submodules
  - `text`
    - `fill`
    - `flip_h`
    - `flip_v`
    - `fill_lines`
    - `flip_lines_h`
    - `flip_lines_v`
    - `rotate`
- Framework
  - `Engine`
  - `Clock`
  - `Screen`
  - `Scene`
- Datastructures
  - `Animation`
  - `AnimationSet`
  - `Hitbox`
- Functions
  - `load_texture`
- Decorators
  - `group`
- Enums
  - `Group`
- Components
  - `TransformComponent`
  - `TextureComponent`
  - `ColorComponent`
  - `AnimatedComponent`
  - `ColliderComponent`
- Nodes
  - `Node`
  - `Node2D`
  - `Camera`
  - `Sprite`
  - `Label`
  - `AnimatedSprite`
- Feature dependent
  - `SimpleMovementComponent` (when using feature `keyboard`/`all`)
"""

from __future__ import annotations as _annotations

__all__ = [
    # Annotations
    "ColorValue",
    "Self",
    # Math
    "lerp",
    "sign",
    "clamp",
    "move_toward",
    "Vec2",
    "Vec2i",
    "Vec3",
    # Submodules
    "text",
    # Framework
    "Engine",
    "Clock",
    "Screen",
    "Scene",
    "AssetLoader",
    # Datastructures
    "Animation",
    "AnimationSet",
    "Hitbox",
    # Functions
    "load_texture",
    # Decorators
    "group",
    # Enums
    "Group",
    # Singletons
    "Time",
    "AssetLoader",
    # Components
    "TransformComponent",
    "TextureComponent",
    "ColorComponent",
    "AnimatedComponent",
    "ColliderComponent",
    # Nodes
    "Node",
    "Node2D",
    "Camera",
    "Sprite",
    "Label",
    "AnimatedSprite",
]

from typing import (
    TYPE_CHECKING as _TYPE_CHECKING,
    TypeAlias as _TypeAlias,
    Literal as _Literal,
    Any as _Any,
    overload as _overload,
    get_args as _get_args,
)

from typing_extensions import assert_never as _assert_never

# Re-exports from `colex`
from colex import ColorValue

# Re-exports from `charz-core`
from charz_core import (
    Self,
    lerp,
    sign,
    clamp,
    move_toward,
    Vec2,
    Vec2i,
    Vec3,
    group,
    TransformComponent,
    Node,
    Node2D,
    Scene,
    Camera,
)

# Exports
from ._engine import Engine
from ._clock import Clock
from ._screen import Screen
from ._time import Time
from ._asset_loader import AssetLoader
from ._grouping import Group
from ._animation import Animation, AnimationSet
from ._components._texture import load_texture, TextureComponent
from ._components._color import ColorComponent
from ._components._animated import AnimatedComponent
from ._components._collision import ColliderComponent, Hitbox
from ._prefabs._sprite import Sprite
from ._prefabs._label import Label
from ._prefabs._animated_sprite import AnimatedSprite
from . import text

# Import to add scene frame tasks
from . import _scene_tasks


# Provide correct completion help - Even if the required feature is not active
if _TYPE_CHECKING:
    from ._components._simple_movement import SimpleMovementComponent

# Lazy exports
# NOTE: Literals and `__getattr__` case branches has to be implemented manually
_LazyNames: _TypeAlias = _Literal["SimpleMovementComponent"]
# NOTE: Add string to `Literal` when a new branch is added
# _LazyNames: _TypeAlias = _Literal["SimpleMovementComponent", "ExampleLazyObject"]
_lazy_object_names = _get_args(_LazyNames)
_lazy_loaded_objects: dict[str, _Any] = {}


# Lazy load to properly load optional dependencies along the standard exports
# NOTE: Using "type: ignore" since it takes multiple branches to work properly
@_overload
def __getattr__(name: _Literal["SimpleMovementComponent"]): ...  # type: ignore  # noqa: ANN202
# @_overload
# def __getattr__(name: _Literal["ExampleLazyObject"]): ...
def __getattr__(name: _LazyNames):
    if name in _lazy_loaded_objects:
        return _lazy_loaded_objects[name]
    elif name in _lazy_object_names:
        # NOTE: Manually add each case branch
        match name:
            case "SimpleMovementComponent":
                from ._components._simple_movement import SimpleMovementComponent  # noqa: PLC0415

                _lazy_loaded_objects[name] = SimpleMovementComponent
                return _lazy_loaded_objects[name]
            case _:
                _assert_never(name)
    raise AttributeError(f"Module '{__name__}' has no attribute '{name}'")



from __future__ import annotations

import types
from functools import partial
from pathlib import Path
from copy import deepcopy
from typing import Iterator

from charz_core import Vec2i, Self

from ._components._texture import load_texture
from ._asset_loader import AssetLoader
from . import text
from ._annotations import Char


class Animation:
    r"""`Animation` dataclass to represent an animation consisting of multiple frames.

    Examples:

    The most common way to create an `Animation`,
    is to load it from disk using the constructor (`__init__`):

    ```python
    from charz import AnimatedSprite, Animation, AnimationSet

    class AnimatedPlayer(AnimatedSprite):
        animations = AnimationSet(
            TestAnimation=Animation("relative/path/to/animation_folder"),
            Run=Animation(
                "animations/player/run",
                fill=True,
                flip_h=True
            ),
        )
        # Setting texture to first frame in animation `Run`
        texture = animations.Run.frames[0]
    ```

    Creating an animation from a list of frames
    using `Animation.from_frames`,
    which are lists of strings representing each frame's texture:

    ```python
    # Extending the previous example...

    class AnimatedPlayer(AnimatedSprite):
        animations = AnimationSet(...)
        ...
        ...
        custom_frames = [
            [
                "  O",
                "/ | \\",
                " / \\",
            ],
            [
                "\\ O /",
                "  |",
                " / \\",
            ],
        ]
        animations.CustomAnimation = Animation.from_frames(custom_frames)
    ```
    """

    __slots__ = ("frames",)

    @classmethod
    def from_frames(
        cls,
        frames: list[list[str]],
        /,
        *,
        reverse: bool = False,
        flip_h: bool = False,
        flip_v: bool = False,
        fill: bool = True,
        fill_char: Char = " ",
        unique: bool = True,
    ) -> Self:
        """Create an `Animation` from a list of frames/textures.

        Args:
            frames (list[list[str]]): List of frames, where each frame is a list of strings.
            reverse (bool, optional): Reverse the order of frames. Defaults to False.
            flip_h (bool, optional): Flip frames horizontally. Defaults to False.
            flip_v (bool, optional): Flip frames vertically. Defaults to False.
            fill (bool, optional): Fill in to make shape of frames rectangular. Defaults to True.
            fill_char (Char, optional): String of length `1` to fill with. Defaults to " ".
            unique (bool, optional): Whether the frames should be unique instances. Defaults to True.

        Returns:
            Self: An instance of `Animation` (or subclass) with the processed frames.

        Raises:
            ValueError: If `fill_char` is not of length `1`.
        """  # noqa: E501
        if len(fill_char) != 1:
            raise ValueError(
                f"Parameter 'fill_char' must be of length 1, got {len(fill_char) = }"
            )
        instance = super().__new__(cls)  # Omit calling `__init__`
        # The negated parameters creates unique list instances,
        # so only copy if they are not present and `unique` is true,
        # else it would be copying an extra time for no reason
        if unique and not reverse and not flip_h and not flip_v and not fill:
            generator = deepcopy(frames)
        else:
            generator = frames

        if fill:  # NOTE: This fill logic has to be before flipping
            generator = map(partial(text.fill_lines, fill_char=fill_char), generator)
        if flip_h:
            generator = map(text.flip_lines_h, generator)
        if flip_v:
            generator = map(text.flip_lines_v, generator)
        if reverse:
            generator = reversed(list(generator))
        instance.frames = list(generator)
        return instance

    def __init__(
        self,
        animation_path: Path | str,
        /,
        *,
        reverse: bool = False,
        flip_h: bool = False,
        flip_v: bool = False,
        fill: bool = True,
        fill_char: Char = " ",
    ) -> None:
        """Load an `Animation` given a path to the folder where the animation is stored.

        Args:
            animation_path (Path | str): Path to folder where animation frames are stored as files.
            flip_h (bool, optional): Flip frames horizontally. Defaults to False.
            flip_v (bool, optional): Flip frames vertically. Defaults to False.
            fill (bool, optional): Fill in to make shape of frames rectangular. Defaults to True.
            fill_char (Char, optional): String of length `1` to fill with. Defaults to " ".
        """  # noqa: E501
        frame_directory = (
            Path.cwd()
            .joinpath(AssetLoader.animation_root)
            .joinpath(animation_path)
            .iterdir()
        )
        generator = map(load_texture, frame_directory)
        if fill:  # NOTE: This fill logic has to be before flipping
            generator = map(partial(text.fill_lines, fill_char=fill_char), generator)
        if flip_h:
            generator = map(text.flip_lines_h, generator)
        if flip_v:
            generator = map(text.flip_lines_v, generator)
        if reverse:
            generator = reversed(list(generator))
        self.frames = list(generator)

    def __repr__(self) -> str:
        # Should never be empty, but if the programmer did it,
        # show empty frame count as 'N/A'
        if not self.frames:
            return f"{self.__class__.__name__}(N/A)"
        longest = 0
        shortest = 0
        tallest = 0
        lowest = 0
        # Variables used in loop to count sizes
        local_longest = 0
        local_shortest = 0
        local_tallest = 0
        local_lowest = 0
        for frame in self.frames:
            # Compare all time best against best results per iteration
            # Allow empty frame and frame with empty lines
            if not frame or not any(frame):
                continue
            local_longest = len(max(frame, key=len))
            longest = max(local_longest, longest)
            local_tallest = len(frame)
            tallest = max(local_tallest, tallest)
            local_shortest = len(min(frame, key=len))
            shortest = min(local_shortest, shortest)
            local_lowest = min(local_lowest, shortest)
        return (
            self.__class__.__name__
            + "("
            + f"{len(self.frames)}"
            + ":{}x{}".format(*self.get_smallest_frame_dimensions())
            + "->{}x{}".format(*self.get_largest_frame_dimensions())
            + ")"
        )

    def get_smallest_frame_dimensions(self) -> Vec2i:
        """Get the smallest frame dimensions in the animation.

        Returns:
            Vec2i: A `Vec2i` object containing the width and height of the smallest frame.
        """
        best_shortest = 0
        best_lowest = 0
        for frame in self.frames:
            if not frame or not any(frame):
                continue
            this_iter_shortest = len(min(frame, key=len))
            best_shortest = min(best_shortest, this_iter_shortest)
            this_iter_lowest = len(frame)
            best_lowest = min(best_lowest, this_iter_lowest)
        return Vec2i(best_shortest, best_lowest)

    def get_largest_frame_dimensions(self) -> Vec2i:
        """Get the largest frame dimensions in the animation.

        Returns:
            Vec2i: A `Vec2i` object containing the width and height of the largest frame.
        """
        best_longest = 0
        best_tallest = 0
        for frame in self.frames:
            if not frame or not any(frame):
                continue
            this_iter_longest = len(max(frame, key=len))
            best_longest = max(best_longest, this_iter_longest)
            this_iter_tallest = len(frame)
            best_tallest = max(best_tallest, this_iter_tallest)
        return Vec2i(best_longest, best_tallest)


class AnimationSet(types.SimpleNamespace):
    """`AnimationSet` dataclass to represent a collection of animations.

    It is subclassed from `types.SimpleNamespace` to allow dynamic attribute access.
    """

    __dict__: dict[str, Animation]

    def __init__(self, **animations: Animation) -> None:
        super().__init__(**animations)

    def __getattribute__(self, name: str) -> Animation:
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Animation) -> None:
        return super().__setattr__(name, value)

    def __iter__(self) -> Iterator[Animation]:
        return iter(self.__dict__.values())


def some(foo: int) -> float:
    """Foobar"""
    return 2



from __future__ import annotations

from colex import ColorValue
from charz_core import Node, Vec2

from .._animation import AnimationSet
from .._components._animated import AnimatedComponent
from ._sprite import Sprite
from .._annotations import Char


class AnimatedSprite(AnimatedComponent, Sprite):
    """`AnimatedSprite` node with multiple textures packed into animations.

    It inherits from `AnimatedComponent` and `Sprite`, allowing it to
    play animations defined in its `AnimationSet, while also being a sprite
    with a texture, position, rotation, and other visual properties.

    Example:

    The paths are relative to the project root,
    unless an animation directory is set in `AssetLoader`:

    ```python
    from charz import AnimatedSprite, AnimationSet, Animation

    class AnimatedGoblin(AnimatedSprite):
        animations = AnimationSet(
            Idle=Animation("path/to/animation_folder/goblin/idle"),
            Attack=Animation("goblin/attack"),
            AttackLeft=Animation("goblin/attack", flip_h=True),
            Flee=Animation("goblin/attack", reverse=True),
        )
        texture = animations.Idle.frames[0]  # Initial texture
        health: int = 10
        damage: int = 2

        def on_attack(self, attacker: Player) -> None:
            self.health -= attacker.sword.damage
            if self.health > 4:
                self.play("Attack")  # Attack back
                attacker.health -= self.damage
            elif self.health > 0
                self.play("Flee")  # Flee on low health
            else:
                self.queue_free()  # Killed
    ```

    `NOTE` See `Animation.__init__` for more options when loading animations.
    """

    def __init__(
        self,
        parent: Node | None = None,
        *,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        texture: list[str] | None = None,
        visible: bool | None = None,
        centered: bool | None = None,
        z_index: int | None = None,
        transparency: Char | None = None,
        color: ColorValue | None = None,
        animations: AnimationSet | None = None,
        repeat: bool | None = None,
    ) -> None:
        Sprite.__init__(
            self,
            parent=parent,
            position=position,
            rotation=rotation,
            top_level=top_level,
            texture=texture,
            visible=visible,
            centered=centered,
            z_index=z_index,
            transparency=transparency,
            color=color,
        )
        if animations is not None:
            self.animations = animations
        if repeat is not None:
            self.repeat = repeat

    def __repr__(self) -> str:
        visual_frame_sizes: list[str] = []
        animation_count = len(self.animations.__dict__)
        if animation_count:
            for animation in self.animations.__dict__.values():
                visual_frame_size = f"{0}x{1}->{2}x{3}".format(
                    *animation.get_smallest_frame_dimensions(),
                    *animation.get_largest_frame_dimensions(),
                )
                visual_frame_sizes.append(visual_frame_size)
        else:
            visual_frame_sizes.append("N/A")

        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{round(self.position, 2)}"
            + f":{round(self.rotation, 2)}R"
            + f":{'{}x{}'.format(*self.get_texture_size())}"
            + f":{repr(self.color)}"
            + f":A={animation_count},"
            + ",".join(visual_frame_sizes)
            + ")"
        )



from __future__ import annotations

import time

from ._non_negative import NonNegative


class Clock:
    """`Clock` class, with delta time calculation.

    Used to sleep for the remaining time of the current frame,
    until a new frame should be processed.

    Examples:

    An instance of `Clock` used by the active `Engine`:

    ```python
    from charz import Engine, Clock

    class DeltaSyncedGame(Engine):
        clock = Clock(fps=12)  # Set frames per second to 12
    ```

    If you don't want the clock to sleep, set `fps` to `0`:

    ```python
    from charz import Engine, Clock

    class SimulatedGame(Engine):
        clock = Clock(fps=0)  # Updates as fast as possible
    ```

    Attributes:
        `fps`: `NonNegative[float]` - Frames per second. If `0`, it will not sleep.
        `delta`: `property[float]` - Read-only attribute for delta time,
            updated on each `tick` call.
    """

    fps = NonNegative[float](0)

    def __init__(self, *, fps: float = 0) -> None:
        """Initialize with optional `fps`.

        `NOTE` When `fps` is set to `0`, it will **not** do any sleeping,
        which means `delta` will be updated and nothing else happens.

        Args:
            fps (float, optional): Frames per second. Defaults to `0`.
        """
        self.fps = fps
        self._delta = 1 / self.fps
        self._last_tick = time.perf_counter()

    def __repr__(self) -> str:
        fps = self.fps  # Assign to temp var to use prettier formatting on next line
        return f"{self.__class__.__name__}({fps=})"

    @property
    def delta(self) -> float:
        """Read-only attribute for delta time.

        Call `tick` to update and mutate properly.
        """
        return self._delta

    def tick(self) -> None:
        """Sleeps for the remaining time to maintain desired `fps`."""
        current_time = time.perf_counter()

        if self.fps == 0:  # Skip sleeping if `.fps` is zero
            self._last_tick = current_time
            return

        target_delta = 1 / self.fps  # Seconds
        elapsed_time = current_time - self._last_tick
        sleep_time = target_delta - elapsed_time
        if sleep_time > 0:
            time.sleep(sleep_time)
            self._last_tick = time.perf_counter()
        else:
            self._last_tick = current_time
        self._delta = max(0, sleep_time)



"""
Text utility module
===================

Utility for flipping characters/lines. Support for rotating characters

Includes
--------

- `fill`
- `flip_h`
- `flip_v`
- `fill_lines`
- `flip_lines_h`
- `flip_lines_v`
- `rotate`
"""

from __future__ import annotations as _annotations

from math import tau as _TAU

from ._annotations import Char


# NOTE: I marked variables with underscores as they are not for export,
#       since this is a publicly exported module of `charz`

# IDEA: Make TextTranslator (as in Transform props),
#       as public export,
#       and wrap methods of 1 default instance with module functions in this file


# Predefined horizontal conversions
_horizontal_conversions: dict[str, str] = {
    "/": "\\",
    "(": ")",
    "[": "]",
    "{": "}",
    ">": "<",
    "´": "`",
    "d": "b",
    "q": "p",
}
# Mirroring `_horizontal_conversions`
# fmt: off
_horizontal_conversions.update({
    value: key
    for key, value in _horizontal_conversions.items()
})
# Unmirrored `_horizontal_conversions` for monodirectional translations
_horizontal_conversions.update({
    "7": "<"
})
# fmt: on
# Predefined vertical conversions
_vertical_conversions: dict[str, str] = {
    "/": "\\",
    ".": "'",
    ",": "`",
    "¯": "_",
    "b": "p",
    "q": "d",
    "w": "m",
    "W": "M",
    "v": "^",
    "V": "A",
}
# Mirroring `_vertical_conversions`
# fmt: off
_vertical_conversions.update({
    value: key
    for key, value in _vertical_conversions.items()
})
# fmt: on
# Unmirrored `_vertical_conversions` for monodirectional translations
# fmt: off
# _vertical_conversions.update({
#     # None for now...
# })
# fmt: on
# Predefined rotational conversions
_rotational_conversions: dict[str, tuple[str, ...]] = {
    "-": ("-", "/", "|", "\\", "-", "/", "|", "\\"),
    ".": (".", "'"),
    "b": ("b", "p", "q", "d"),
    "9": ("9", "6"),
}

# Creating mirrored {char: variants} pairs,
# for pairs already defined in `_rotational_conversions`

# Mirror `_rotational_conversions` (adds variants as their own keys)
for _options in list(_rotational_conversions.values()):
    for _idx, _value in enumerate(_options):
        if _idx == 0:  # Skip existing initial pair
            continue
        if _value in _rotational_conversions:
            # Some char variants occur multiple times
            # NOTE: Define `_rotational_conversions` keys using "resting variants"
            continue
        _before = _options[:_idx]
        _after = _options[_idx:]
        _new_values = (*_after, *_before)
        _rotational_conversions[_value] = _new_values
# Unmirrored `_rotational_conversions` for spesific lookup translations
# fmt: off
# _rotational_conversions.update({
#     # None for now...
# })
# fmt: on


def fill(line: str, *, width: int, fill_char: Char = " ") -> str:
    """Fill a single left-justified line with a string of length `1`.

    Args:
        line (str): Line to be filled.
        width (int): Maximum width of output string.
        fill_char (Char, optional): String of length `1` to fill line. Defaults to " ".

    Returns:
        str: Line filled with fill character.
    """
    return line.ljust(width, fill_char)


def flip_h(line: str, /) -> str:
    """Flip a single line horizontally.

    Also works with a single character.

    Args:
        line (list[str]): Content to be flipped.

    Returns:
        list[str]: Flipped line or character.
    """
    return "".join(_horizontal_conversions.get(char, char) for char in reversed(line))


def flip_v(line: str, /) -> str:
    """Flip a single line vertically.

    Also works with a single character.

    Args:
        line (list[str]): Content to be flipped.

    Returns:
        list[str]: Flipped line or character.
    """
    return "".join(_vertical_conversions.get(char, char) for char in line)


def fill_lines(lines: list[str], *, fill_char: Char = " ") -> list[str]:
    """Fill lines with fill character, based on longest line.

    Usefull for filling textures, so that it gets a nice rectangular shape.
    Good for centering and flipping textures.

    Args:
        lines (list[str]): Lines to be filled.
        fill_char (Char, optional): String of length `1` to fill line. Defaults to " ".

    Returns:
        list[str]: Rectangular filled lines.
    """
    if not lines:  # Allow when there are no lines
        return []  # But still return unique list
    longest = len(max(lines, key=len))
    return [fill(line, width=longest, fill_char=fill_char) for line in lines]


def flip_lines_h(lines: list[str], /) -> list[str]:
    """Flip lines horizontally.

    Usefull for flipping textures.

    Args:
        lines (list[str]): Lines of strings or texture

    Returns:
        list[str]: Flipped content
    """
    return [flip_h(line) for line in lines]


def flip_lines_v(lines: list[str], /) -> list[str]:
    """Flip lines vertically.

    Usefull for flipping textures.

    Args:
        lines (list[str]): Lines of strings or texture.

    Returns:
        list[str]: Flipped content.
    """
    return [flip_v(line) for line in reversed(lines)]


def rotate(char: Char, /, angle: float) -> str:
    """Return symbol when rotated by angle counter clockwise.

    Args:
        char (Char): String of length `1` to rotate.
        angle (float): Counter clockwise rotation in radians.

    Returns:
        str: Rotated character or original character.
    """
    if char in _rotational_conversions:
        sector_count = len(_rotational_conversions[char])
        sector_rads = _TAU / sector_count
        half_sector_rads = sector_rads / 2
        total_rads = (angle + half_sector_rads) % _TAU
        index = int(total_rads / sector_rads) % sector_count
        return _rotational_conversions[char][index]
    return char



from __future__ import annotations

import os
import sys
from math import cos, sin, floor
from enum import Enum, unique, auto
from typing import Sequence

from colex import ColorValue, RESET
from charz_core import Scene, Camera, TransformComponent, Vec2i

from . import text
from ._components._texture import TextureComponent
from ._grouping import Group
from ._annotations import TextureNode, FileLike, Renderable, Char


@unique
class ConsoleCode(str, Enum):
    CLEAR = "\x1b[2J\x1b[H"


@unique
class CursorCode(str, Enum):
    HIDE = "\x1b[?25l"
    SHOW = "\x1b[?25h"


@unique
class ColorChoice(Enum):
    AUTO = auto()
    ALWAYS = auto()
    NEVER = auto()


class ScreenClassProperties(type):
    """Workaround to add class properties to `Screen`."""

    COLOR_CHOICE_AUTO = ColorChoice.AUTO
    COLOR_CHOICE_ALWAYS = ColorChoice.ALWAYS
    COLOR_CHOICE_NEVER = ColorChoice.NEVER


class Screen(metaclass=ScreenClassProperties):
    """`Screen` class, representing a virtual screen for rendering `ASCII` frames.

    An instance of `Screen` is used by the active `Engine`.

    `NOTE` Attribute `stream` is defined at *class level*,
    which means **all instances will share the same reference**,
    unless explicitly overridden.

    Example:

    ```python
    from charz import Engine, Screen

    class MyGame(Engine):
        screen = Screen(
            width=80,
            height=24,
            color_choice=Screen.COLOR_CHOICE_AUTO,
        )
    ```

    Attributes:
        `stream`: `FileLike[str]` - Output stream written to.
            Defaults to `sys.stdout`.
        `buffer`: `list[list[tuple[Char, ColorValue | None]]]` - Screen buffer,
            where each pixel is stored in a 2D `list`,
            and each pixel is a `tuple` pair of visual character and optional color.
        `width`: `NonNegative[int]` - Viewport width in character pixels.
        `height`: `NonNegative[int]` - Viewport height in character pixels.
        `size`: `property[Vec2i]` - Read-only getter,
            which packs `width` and `height` into a `Vec2i` instance.
        `auto_resize`: `property[bool]` - Whether to use terminal size as viewport size.
        `intial_clear`: `bool` - Whether to clear terminal on startup.
        `final_clear`: `bool` - Whether to clear screen on cleanup.
        `hide_cursor`: `bool` - Whether to hide cursor.
        `transparency_fill`: `Char` - Character used for transparent pixels..
        `color_choice`: `ColorChoice` - How colors are handled.
        `margin_right`: `int` - Margin on right side to not draw on.
        `margin_bottom`: `int` - Margin under to not draw on.

    Hooks:
        `on_startup`
        `on_cleanup`

    Methods:
        `is_using_ansi`
        `get_actual_size`
        `reset_buffer`
        `render_all`
        `show`
        `refresh`
    """

    stream: FileLike[str] = sys.stdout
    buffer: list[list[tuple[Char, ColorValue | None]]]

    def __init__(
        self,
        width: int = 16,
        height: int = 12,
        *,
        auto_resize: bool = False,
        initial_clear: bool = True,
        final_clear: bool = True,
        hide_cursor: bool = True,
        transparency_fill: Char = " ",
        color_choice: ColorChoice = ColorChoice.AUTO,
        stream: FileLike[str] | None = None,
        margin_right: int = 1,
        margin_bottom: int = 1,
    ) -> None:
        """Initialize screen with given width and height.

        Args:
            width (NonNegative[int]): Viewport width of the screen in characters.
            height (NonNegative[int]): Viewport height of the screen in characters.
            auto_resize (bool): Whether to automatically resize the screen,
                based on terminal size. Defaults to `False`.
            initial_clear (bool): Whether to clear the screen on startup.
                Defaults to `True`.
            final_clear (bool): Whether to clear the screen on cleanup.
                Defaults to `True`.
            hide_cursor (bool): Whether to hide the cursor on startup.
                Defaults to `True`.
            transparency_fill (Char): Character used for transparent pixels.
                Defaults to `" "`.
            color_choice (ColorChoice): How colors are handled.
                Defaults to `Screen.COLOR_CHOICE_AUTO`.
            stream (FileLike[str] | None): Output stream.
                Defaults to `sys.stdout`.
            margin_right (int): Right margin in characters.
                Defaults to `1`.
            margin_bottom (int): Bottom margin in characters.
                Defaults to `1`.

        Raises:
            ValueError: If `transparency_fill` is not a `str` of length `1`.
        """
        if len(transparency_fill) != 1:
            raise ValueError(
                f"String length not equal to 1, got {len(transparency_fill) = }"
            )
        self.width = width
        self.height = height
        self.color_choice = color_choice
        if stream is not None:
            self.stream = stream
            # NOTE: Uses class variable `Screen.stream` by default
        self.margin_right = margin_right
        self.margin_bottom = margin_bottom
        self._auto_resize = auto_resize
        self.initial_clear = initial_clear
        self.final_clear = final_clear
        self.hide_cursor = hide_cursor
        self._resize_if_necessary()
        self.transparency_fill = transparency_fill
        self.buffer = []
        self.reset_buffer()  # For populating the screen buffer

    def on_startup(self) -> None:
        """Startup hook.

        Called when the screen is being activated.
        The logic is seperated into this method,
        as only 1 screen (which normally uses `sys.stdout`) can be active at a time.

        Multiple screens can be used at the same time,
        as long as they use a different type of filehandle (like sockets or files),
        though this is not recommended.
        """
        if self.is_using_ansi():
            if self.initial_clear:
                self.stream.write(ConsoleCode.CLEAR)
                self.stream.flush()
            if self.hide_cursor:
                self.stream.write(CursorCode.HIDE)
                self.stream.flush()

    def on_cleanup(self) -> None:
        """Cleanup hook.

        Called when the screen is being deactivated.
        The logic is seperated into this method,
        as only 1 screen (which normally uses `sys.stdout`) can be active at a time.
        """
        if self.hide_cursor and self.is_using_ansi():
            self.stream.write(CursorCode.SHOW)
            self.stream.flush()
        if self.final_clear:
            old_fill = self.transparency_fill
            self.transparency_fill = " "
            self.reset_buffer()
            self.show()
            self.transparency_fill = old_fill

    @property
    def auto_resize(self) -> bool:
        """Whether the screen automatically resizes based on terminal size.

        Returns:
            bool: `True` if auto-resizing is enabled, `False` otherwise.
        """
        return self._auto_resize

    @auto_resize.setter
    def auto_resize(self, state: bool) -> None:
        """Set whether the screen should automatically resize based on terminal size.

        Args:
            state (bool): `True` to enable auto-resizing, `False` to disable.
        """
        self._auto_resize = state
        self._resize_if_necessary()

    def _resize_if_necessary(self) -> None:
        """Resize the screen if requirements are met or circumstances have changed.

        This method checks if the screen should be resized based on the
        `auto_resize` property and the current terminal size.
        If `auto_resize` is `True`, it attempts to get the terminal size
        using `os.get_terminal_size(...)`.
        If successful, it updates the screen dimensions accordingly.

        `NOTE` Does not mutate screen `buffer`.
        """
        if self.auto_resize:
            try:
                fileno = self.stream.fileno()
            except (ValueError, OSError):
                # Do not resize if not proper `.stream.fileno()` is available,
                # like `io.StringIO.fileno()`
                return
            try:
                terminal_size = os.get_terminal_size(fileno)
            except (ValueError, OSError):
                return
            self.width = terminal_size.columns - self.margin_right
            self.height = terminal_size.lines - self.margin_bottom

    @property
    def size(self) -> Vec2i:
        """Get the size of the screen as a `Vec2i`.

        Returns:
            Vec2i: Width and height of the screen, represented by a `Vec2i`.
        """
        return Vec2i(self.width, self.height)

    @size.setter
    def size(self, size: Vec2i) -> None:
        """Set the size of the screen.

        Args:
            size (Vec2i): Width and height, represented by a `Vec2i`.

        Raises:
            ValueError: If the size `Vec2i` could not be unpacked into `2`x`int`.
        """
        width, height = size
        if not isinstance(width, int):
            raise ValueError(f"Width cannot be of type '{type(size)}', expected 'int'")
        if not isinstance(height, int):
            raise ValueError(f"Height cannot be of type '{type(size)}', expected 'int'")
        self.width = width
        self.height = height
        self._resize_if_necessary()

    def is_using_ansi(self) -> bool:
        """Return whether its using ANSI escape and color codes.

        Checks first `.color_choice`. Returns `True` if set to `ALWAYS`,
        and `False` if set to `NEVER`.
        If set to `AUTO`, check whether a `tty` is detected.

        Returns:
            bool: `True` if using ANSI codes, `False` otherwise.
        """
        if self.color_choice is ColorChoice.ALWAYS:
            return True
        try:
            fileno = self.stream.fileno()
        except (ValueError, OSError):
            is_a_tty = False
        else:
            try:
                is_a_tty = os.isatty(fileno)
            except OSError:
                is_a_tty = False
        # Returns `False` if not a TTY or is `ColorChoice.NEVER`
        return self.color_choice is ColorChoice.AUTO and is_a_tty

    def get_actual_size(self) -> Vec2i:
        """Get the actual size of the screen based on terminal size.

        The `width` and `height` of the screen are just theoretical maximums,
        though the real terminal might be smaller than these values.
        It also takes into account the right and bottom margins,
        which are nice if jittering occurs because of not accurate values
        reported by `os.get_terminal_size(...)`.
        If `stream` is not set to `sys.stdout`, it will return `width` and `height`.

        Returns:
            Vec2i: Actual size of the screen, adjusted for terminal size and margins.
        """
        try:
            fileno = self.stream.fileno()
        except (ValueError, OSError):
            return self.size.copy()
        try:
            terminal_size = os.get_terminal_size(fileno)
        except (ValueError, OSError):
            return self.size.copy()
        actual_width = min(self.width, terminal_size.columns - self.margin_right)
        actual_height = min(self.height, terminal_size.lines - self.margin_bottom)
        return Vec2i(actual_width, actual_height)

    def reset_buffer(self) -> None:
        """Clear the screen `buffer`.

        It will fill the buffer with the transparency fill character,
        as well as `None` for the color, per "pixel".
        """
        self.buffer = [
            # Pair structure: (char, color)
            [(self.transparency_fill, None) for _ in range(self.width)]
            for _ in range(self.height)
        ]

    def render_all(self, nodes: Sequence[Renderable], /) -> None:
        """Render all nodes provided to the screen buffer.

        Args:
            nodes (Sequence[Renderable]): Sequence of nodes with `TextureComponent`.

        Raises:
            ValueError: If a any node has an invalid transparency character length,
                which is not equal to `1`.
        """
        nodes_sorted_by_z_index = sorted(nodes, key=lambda node: node.z_index)

        # Include half size of camera parent when including size
        viewport_global_position = Camera.current.global_position

        if Camera.current.mode & Camera.MODE_INCLUDE_SIZE and isinstance(
            Camera.current.parent, TextureComponent
        ):
            # Adds half of camera's parent's texture size
            viewport_global_position += Camera.current.parent.get_texture_size() / 2

        # Determine whether to use use the parent of current camera
        # or its parent as anchor for viewport
        anchor = Camera.current
        if not Camera.current.top_level and isinstance(
            Camera.current.parent, TransformComponent
        ):
            anchor = Camera.current.parent

        for node in nodes_sorted_by_z_index:
            if not node.is_globally_visible():
                continue

            if node.transparency is not None and len(node.transparency) != 1:
                raise ValueError(
                    f"Node {node} has invalid transparency character length:"
                    f"{len(node.transparency)} != 1"
                )

            # Cache and lookup node properties/attributes
            node_global_position = node.global_position
            node_global_rotation = node.global_rotation
            node_color: ColorValue | None = getattr(node, "color")  # noqa: B009

            relative_position = node_global_position - anchor.global_position

            if Camera.current.mode & Camera.MODE_CENTERED:
                relative_position.x += self.width / 2
                relative_position.y += self.height / 2

            # Offset from centering
            offset_x = 0
            offset_y = 0
            if node.centered:
                node_texture_size = node.get_texture_size()
                offset_x = node_texture_size.x / 2
                offset_y = node_texture_size.y / 2

            # TODO: Improve performance by using simpler solution when no rotation
            #     - Iterate over each cell, for each row
            for h, row in enumerate(node.texture):
                for w, char in enumerate(row):
                    if char == node.transparency:
                        continue

                    # Adjust starting point based on centering
                    x_diff = w - offset_x
                    y_diff = h - offset_y

                    # Apply rotation using upper-left as the origin
                    # NOTE: `-node_global_rotation` means counter clockwise
                    # Cache using variable
                    neg_node_global_rotation = -node_global_rotation
                    rotated_x = (
                        cos(neg_node_global_rotation) * x_diff
                        - sin(neg_node_global_rotation) * y_diff
                    )
                    rotated_y = (
                        sin(neg_node_global_rotation) * x_diff
                        + cos(neg_node_global_rotation) * y_diff
                    )

                    # Translate to final screen position
                    final_x = relative_position.x + rotated_x
                    final_y = relative_position.y + rotated_y

                    # Apply horizontal index snap
                    char_index = floor(final_x)
                    # Do horizontal boundary check
                    if char_index < 0 or char_index >= self.width:
                        continue

                    # Apply vertical index snap
                    row_index = floor(final_y)
                    # Do vertical boundary check
                    if row_index < 0 or row_index >= self.height:
                        continue

                    # Insert rotated char into screen buffer
                    rotated_char = text.rotate(char, node_global_rotation)
                    self.buffer[row_index][char_index] = (rotated_char, node_color)

    def show(self) -> None:
        """Show content of screen buffer.

        This will print the formatted frame to the terminal,
        if `stream` is set to `sys.stdout`.
        """
        actual_size = self.get_actual_size()
        # Construct frame from screen buffer
        out = ""
        is_using_ansi = self.is_using_ansi()
        for lino, row in enumerate(self.buffer[: actual_size.y], start=1):
            for char, color in row[: actual_size.x]:
                if is_using_ansi:
                    if color is None:
                        out += RESET + char
                    else:
                        out += RESET + color + char
                else:
                    out += char
            if lino != actual_size.y:  # Not at end
                out += "\n"
        if is_using_ansi:
            out += RESET
            cursor_move_code = f"\x1b[{actual_size.y - 1}A" + "\r"
            out += cursor_move_code
        # Write and flush
        self.stream.write(out)
        self.stream.flush()

    def refresh(self) -> None:
        """Refresh the screen, by performing multiple steps.

        The steps are:
        1. Resize screen if necessary.
        2. Reset screen buffer.
        3. Render all texture nodes in current scene.
        4. Show rendered content in terminal.
        """
        self._resize_if_necessary()
        self.reset_buffer()
        # NOTE: Uses underlying `list` because it's faster than `tuple` when copying
        texture_nodes = Scene.current.get_group_members(
            Group.TEXTURE,
            type_hint=TextureNode,
        )
        self.render_all(texture_nodes)
        self.show()



from __future__ import annotations

from pathlib import Path
from typing import Any, NoReturn, final


class AssetLoaderClassProperties(type):
    """Workaround to add class properties to `AssetLoader`."""

    _texture_root: Path = Path.cwd()
    _animation_root: Path = Path.cwd()

    @property
    def texture_root(cls) -> Path:
        return cls._texture_root

    @texture_root.setter
    def texture_root(cls, new_path: Path | str) -> None:
        cls._texture_root = Path(new_path)
        if not cls._texture_root.exists():
            raise ValueError("Invalid sprite root folder path")

    @property
    def animation_root(cls) -> Path:
        return cls._animation_root

    @animation_root.setter
    def animation_root(cls, new_path: Path | str) -> None:
        cls._animation_root = Path(new_path)
        if not cls.animation_root.exists():
            raise ValueError("Invalid animation root folder path")


@final
class AssetLoader(metaclass=AssetLoaderClassProperties):
    """`AssetLoader` is a configuration namespace for loading assets.

    Paths fields is of type `pathlib.Path`,
    and use setters that allow passing either `pathlib.Path` or `str` paths.

    `NOTE` Variables have to be set **before** importing *local files* in your project.
    It is typical to use `load_texture` or create `Animation` instances
    in the class definition when subclassing `Sprite`/`AnimatedSprite`,
    which means these configuration variables has to be set before being used.

    Example:

    Configuring `AssetLoader` attributes the correct way:

    ```python
    from charz import ..., AssetLoader, ...

    AssetLoader.texture_root = "src/sprites"
    AssetLoader.animation_root = "src/animations"

    from .my_custom_node import ...
    ```

    Attributes:
        `texture_root`: `Path` - Relative path to texture/sprites folder.
        `animation_root`: `Path` - Relative path to animations folder.
    """

    def __new__(cls, *_args: Any, **_kwargs: Any) -> NoReturn:
        raise RuntimeError(f"{cls.__name__} cannot be instantiated")



from __future__ import annotations

from typing import NoReturn, Any, final

from ._non_negative import NonNegative


@final
class Time:
    """`Time` is a class namespace used to store delta time.

    `Time.delta` is computed by `Clock`, handled by `Engine` frame task.
    `Time.delta` is usually used in `Node.update`,
    for syncing movement with real time seconds.

    Example:

    ```python
    from charz import Sprite, Time

    class GravityBox(Sprite):
        _GRAVITY_STRENGTH: float = 20  # Positive value means falling down
        _speed_y: float = 0
        texture = [
            "####",
            "####",
            "####",
        ]

        def update(self) -> None:
            # m/s^2 * s = m/s
            self._speed_y += self.GRAVITY_STRENGTH * Time.delta
            # m/s * s = m
            self.position.y += self._speed_y * Time.delta
    ```
    """

    delta = NonNegative[float](0)

    def __new__(cls, *_args: Any, **_kwargs: Any) -> NoReturn:
        raise RuntimeError(f"{cls.__name__} cannot be instantiated")



from __future__ import annotations

import charz_core

from ._clock import Clock
from ._screen import Screen
from ._time import Time


class Engine(charz_core.Engine):
    """Extended `Engine` for managing scenes, clock, and screen.

    It extends the core `Engine` and adds frame tasks
    for managing the clock and screen refresh.

    The run method is wrapped; to set up the screen,
    along side assigning the initial delta time to `Time.delta`.

    Attributes:
        clock (Clock): The clock instance for managing frame timing.
        screen (Screen): The screen instance for rendering output.

    Example:

    ```python
    from charz import Engine, Screen, Clock

    class MyGame(Engine):
        clock = Clock(fps=8)  # Set frames per second
        screen = Screen(
            width=80,
            height=24,
            initial_clear=True,
        )
    ```

    Could also use a custom `Screen` subclass (`charz_rust.RustScreen`),
    that was implemented in `Rust` for better performance
    """

    clock: Clock = Clock(fps=16)
    screen: Screen = Screen()

    def run(self) -> None:  # Extended main loop function
        """Run app/game, which will start the main loop.

        The loop will run until `is_running` is set to `False.

        This function is also responsible for setting up the `screen`,
        which given as an overridden class attribute of `Engine` subclass.
        """
        Time.delta = self.clock.delta
        # Handle special ANSI codes to setup
        self.screen.on_startup()
        super().run()
        # Run cleanup function to clear output screen
        self.screen.on_cleanup()


# Define additional frame tasks


def refresh_screen(engine: Engine) -> None:
    """Call `refresh` on the screen to update the display."""
    engine.screen.refresh()


def tick_clock(engine: Engine) -> None:
    """Tick the clock to update the delta time."""
    engine.clock.tick()
    Time.delta = engine.clock.delta


# Register additional frame tasks
# Priorities are chosen with enough room to insert many more tasks in between
Engine.frame_tasks[80] = refresh_screen
Engine.frame_tasks[70] = tick_clock



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import Any

from charz_core import Vec2i, Self, group

from .. import text
from .._asset_loader import AssetLoader
from .._grouping import Group
from .._annotations import Char


def load_texture(
    texture_path: Path | str,
    /,
    *,
    flip_h: bool = False,
    flip_v: bool = False,
    fill: bool = True,
    fill_char: Char = " ",
) -> list[str]:
    """Load texture from file.

    `NOTE` `AssetLoader.texture_root` will be prepended to `texture_path`.

    Args:
        texture_path (Path | str): Path to file with texture.
        flip_h (bool, optional): Flip horizontally. Defaults to `False`.
        flip_v (bool, optional): Flip vertically. Defaults to `False`.
        fill (bool, optional): Fill in to make shape rectangular. Defaults to `True`.
        fill_char (Char, optional): Filler string of length `1` to use. Defaults to `" "`.

    Returns:
        list[str]: Loaded texture.

    Raises:
        ValueError: If `fill_char` is not of length `1`.
    """
    if len(fill_char) != 1:
        raise ValueError(
            f"Parameter 'fill_char' must of length 1, got {len(fill_char) = }"
        )
    # fmt: off
    file = (
        Path.cwd()
        .joinpath(AssetLoader.texture_root)
        .joinpath(texture_path)
    )
    # fmt: on
    content = file.read_text(encoding="utf-8")
    texture = content.splitlines()
    if fill:  # NOTE: This fill logic has to be before flipping
        texture = text.fill_lines(texture, fill_char=fill_char)
    if flip_h:
        texture = text.flip_lines_h(texture)
    if flip_v:
        texture = text.flip_lines_v(texture)
    return texture


@group(Group.TEXTURE)
class TextureComponent:  # Component (mixin class)
    """`TextureComponent` mixin class for node.

    Attributes:
        `texture`: `list[str]` - The texture data as a list of lines.
        `unique_texture`: `bool` - Whether the texture is unique per instance.
        `visible`: `bool` - Visibility state of the node.
        `centered`: `bool` - Whether the texture is centered.
        `z_index`: `int` - Z-order for rendering.
        `transparency`: `Char | None` - Character used to signal transparency.

    Methods:
        `hide`
        `show`
        `is_globally_visible`
        `get_texture_size`
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        if (class_texture := getattr(instance, "texture", None)) is not None:
            if instance.unique_texture:
                instance.texture = deepcopy(class_texture)
            else:
                instance.texture = class_texture
        else:
            instance.texture = []
        return instance

    texture: list[str]
    unique_texture: bool = True
    visible: bool = True
    centered: bool = False
    z_index: int = 0
    transparency: Char | None = None

    def with_texture(self, texture_or_line: list[str] | str | Char, /) -> Self:
        """Chained method to set the texture of the node.

        If a string is provided, it is treated as a single line texture.

        Args:
            texture_or_line (list[str] | str | Char):
                Texture data as a list of lines, a single line string, or a character.

        Returns:
            Self: Same node instance.
        """
        if isinstance(texture_or_line, str):
            self.texture = [texture_or_line]
            return self
        self.texture = texture_or_line
        return self

    def with_unique_texture(self) -> Self:
        """Chained method to create unique copy of `texture`, and use that.

        Uses `deepcopy` to create the copy.

        Returns:
            Self: Same node instance with a unique texture copy.
        """
        self.texture = deepcopy(self.texture)
        return self

    def with_visibility(self, state: bool = True, /) -> Self:
        """Chained method to set the visibility of the node.

        Args:
            state (bool, optional): Visibility state. Defaults to True.

        Returns:
            Self: Same node instance.
        """
        self.visible = state
        return self

    def with_centering(self, state: bool = True, /) -> Self:
        """Chained method to set whether the texture is centered.

        Args:
            state (bool, optional): Centering state. Defaults to True.

        Returns:
            Self: Same node instance.
        """
        self.centered = state
        return self

    def with_z_index(self, z_index: int, /) -> Self:
        """Chained method to set the z-index for rendering.

        Args:
            z_index (int): Z-index value.

        Returns:
            Self: Same node instance.
        """
        self.z_index = z_index
        return self

    def with_transparency(self, char: Char | None, /) -> Self:
        """Chained method to set the transparency character.

        Uses a string of length `1` as transparency character.
        If `None` is passed, no transparency is applied,
        which means strings with spaces will be rendered on top
        of other nodes with texture
        (as long as it has a greater z-index or the node is newer).

        Args:
            char (Char | None): Transparency character or `None`.

        Returns:
            Self: Same node instance.
        """
        self.transparency = char
        return self

    def hide(self) -> None:
        """Set the node to be hidden."""
        self.visible = False

    def show(self) -> None:
        """Set the node to be visible."""
        self.visible = True

    def is_globally_visible(self) -> bool:
        """Check whether the node and its ancestors are visible.

        Returns:
            bool: Global visibility.
        """
        if not self.visible:
            return False
        parent = self.parent  # type: ignore
        while parent is not None:
            if not isinstance(parent, TextureComponent):
                return True
            if not parent.visible:
                return False
            parent = parent.parent  # type: ignore
        return True

    def get_texture_size(self) -> Vec2i:
        """Get the size of the texture.

        Computed in O(n*m), where n is the number of lines
        and m is the length of the longest line.

        Returns:
            Vec2i: Texture size.
        """
        if not self.texture:
            return Vec2i.ZERO
        return Vec2i(
            len(max(self.texture, key=len)),  # Length of longest line
            len(self.texture),  # Line count
        )



from __future__ import annotations

from colex import ColorValue
from charz_core import Node, Vec2, Self

from ._sprite import Sprite
from .._annotations import Char


class Label(Sprite):
    r"""`Label` node to simplify displaying text.

    It extends `Sprite` and allows for easy text manipulation.
    The text is stored in the `texture` attribute, which is a list of strings.
    By using the `text` property, you can easily set and get the text.

    How tabs and newlines are handled can be customized with the following attributes:
    - `newline`: The character used to separate lines in the text.
        Default: `"\n"`
    - `tab_size`: The number of characters to use for each tab.
        Default: `4`
    - `tab_char`: The character to represent a tab.
        Default: `"\t"`
    - `tab_fill`: The character to use for filling in tabs.
        Default: `" "`

    Examples:

    Creating a `Label` instance inside a `Scene.__init__`:

    ```python
    import colex
    from charz import Scene, Label

    class MenuPage(Scene):
        def __init__(self) -> None:
            self.label = Label(
                text="Welcome to the game!\nPress [Enter] to start.",
                color=colex.BLUE,
                centered=True,
            )
    ```

    Creating the `Label` at top level in `Engine.__init__`:

    ```python
    import colex
    from charz import Engine, Label

    class MyGame(Engine):
        def __init__(self) -> None:
            self.label = Label(
                text="Welcome to the game!\nPress [Enter] to start.",
                color=colex.BLUE,
                centered=True,
            )
    ...
    ...
    game = MyGame()
    game.run()
    ```
    """

    newline: Char = "\n"
    tab_size: int = 4
    tab_char: Char = "\t"
    tab_fill: Char = " "

    def __init__(
        self,
        parent: Node | None = None,
        *,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        texture: list[str] | None = None,
        visible: bool | None = None,
        centered: bool | None = None,
        z_index: int | None = None,
        transparency: Char | None = None,
        color: ColorValue | None = None,
        text: str | None = None,
        newline: Char | None = None,
        tab_size: int | None = None,
        tab_char: Char | None = None,
        tab_fill: Char | None = None,
    ) -> None:
        Sprite.__init__(
            self,
            parent=parent,
            position=position,
            rotation=rotation,
            top_level=top_level,
            texture=texture,
            visible=visible,
            centered=centered,
            z_index=z_index,
            transparency=transparency,
            color=color,
        )
        if text is not None:
            self.text = text
        if newline is not None:
            self.newline = newline
        if tab_size is not None:
            self.tab_size = tab_size
        if tab_char is not None:
            self.tab_char = tab_char
        if tab_fill is not None:
            self.tab_fill = tab_fill

    def with_newline(self, newline: Char, /) -> Self:
        """Chained method to set the newline character.

        Args:
            newline (Char): Newline character to use.

        Returns:
            Self: Same node instance.
        """
        self.newline = newline
        return self

    def with_tab_size(self, tab_size: int, /) -> Self:
        """Chained method to set the tab size.

        Args:
            tab_size (int): Tab size to use.

        Returns:
            Self: Same node instance.
        """
        self.tab_size = tab_size
        return self

    def with_tab_char(self, tab_char: Char, /) -> Self:
        """Chained method to set the tab character.

        Args:
            tab_char (Char): Tab character to use.

        Returns:
            Self: Same node instance.
        """
        self.tab_char = tab_char
        return self

    def with_tab_fill(self, tab_fill: Char, /) -> Self:
        """Chained method to set the tab fill character.

        Args:
            tab_fill (Char): Tab fill character to use.

        Returns:
            Self: Same node instance.
        """
        self.tab_fill = tab_fill
        return self

    def with_text(self, text: str, /) -> Self:
        """Chained method to set the text of the label.

        Args:
            text (str): Text to set as texture.

        Returns:
            Self: Same node instance.
        """
        self.text = text
        return self

    @property
    def text(self) -> str:
        """Get the text of the label.

        This replaces tabs with the node's fill character.

        Returns:
            str: The text of the label with tabs replaced by the fill character.
        """
        joined_lines = self.newline.join(self.texture)
        return joined_lines.replace(self.tab_fill * self.tab_size, self.tab_char)

    @text.setter
    def text(self, value: str) -> None:
        """Set the text of the label.

        This splits the text into lines and replaces tabs with the node's fill character.

        Args:
            value (str): The text to set as the label's texture.
        """
        tab_replaced = self.newline.replace(self.tab_char, self.tab_fill * self.tab_size)
        self.texture = value.split(tab_replaced)



from __future__ import annotations

from typing import Generic, Any

from ._annotations import Number


class NonNegative(Generic[Number]):
    """`NonNegative` descriptor for attributes that must be non-negative.

    It is generic, with constraints for `int` and `float` types.

    `NOTE` This is only tested on class level attributes, not instance attributes.

    Example:

    This ensures that `delta` is always a non-negative float,
    and raises an error if a negative value is assigned.

    ```python
    class Clock:
        delta = NonNegative[float](0)
    ```

    Raises:
        TypeError: If the value is not `int` or `float`.
        ValueError: If the value is negative.
    """

    def __init__(self, value: Number, /) -> None:
        """Initialize the descriptor with a non-negative value.

        Args:
            value (Number): The initial value, must be non-negative.
        """
        if not isinstance(value, Number.__constraints__):
            raise TypeError(
                f"Attribute '{self._name[1:]}' must be {self._valid_types_message()}"
            )
        if value < 0:
            raise ValueError(f"Attribute '{self._name[1:]}' must be non-negative")
        self.value = value

    def __set_name__(self, _owner: type, name: str) -> None:
        self._name = f"_{name}"

    def __get__(self, instance: Any, _owner: type) -> Number:  # noqa: ANN401
        return getattr(instance, self._name, self.value)

    def __set__(self, instance: Any, value: Number) -> None:  # noqa: ANN401
        if not isinstance(value, Number.__constraints__):
            raise TypeError(
                f"Attribute '{self._name[1:]}' must be {self._valid_types_message()}"
            )
        if value < 0:
            raise ValueError(f"Attribute '{self._name[1:]}' must be non-negative")
        setattr(instance, self._name, value)

    def _valid_types_message(self) -> str:
        return " or ".join(
            f"'{constaint.__name__}'" for constaint in Number.__constraints__
        )



from __future__ import annotations

import charz_core
from charz_core import Node, Vec2
from colex import ColorValue

from .._components._texture import TextureComponent
from .._components._color import ColorComponent
from .._annotations import Char


class Sprite(ColorComponent, TextureComponent, charz_core.Node2D):
    r"""`Sprite` node to represent a 2D sprite with texture and color.

    This is the base class for every node that has a texture and color in 2D space.
    Most of the visual nodes will inherit from this class.

    Example:

    Subclassing `Sprite` and overriding class attributes to customize look:

    ```python
    import colex
    from charz import Sprite

    class CustomSprite(Sprite):
        color = colex.RED
        transparency = " "
        centered = True
        texture = [
            "  O",
            "/ | \\",
            " / \\",
        ]
    ```
    """

    def __init__(
        self,
        parent: Node | None = None,
        *,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        texture: list[str] | None = None,
        visible: bool | None = None,
        centered: bool | None = None,
        z_index: int | None = None,
        transparency: Char | None = None,
        color: ColorValue | None = None,
    ) -> None:
        charz_core.Node2D.__init__(
            self,
            parent=parent,
            position=position,
            rotation=rotation,
            top_level=top_level,
        )
        if texture is not None:
            self.texture = texture
        if visible is not None:
            self.visible = visible
        if centered is not None:
            self.centered = centered
        if z_index is not None:
            self.z_index = z_index
        if transparency is not None:
            self.transparency = transparency
        if color is not None:
            self.color = color

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{round(self.position, 2)}"
            + f":{round(self.rotation, 2)}R"
            + f":{'{}x{}'.format(*self.get_texture_size())}"
            + f":{repr(self.color)}"
            + ")"
        )



from __future__ import annotations

from copy import deepcopy
from enum import Enum, unique, auto
from typing import Any

from charz_core import Self, group, clamp

from .._animation import AnimationSet, Animation
from .._grouping import Group


@unique
class PlaybackDirection(Enum):
    FORAWRD = auto()
    BACKWARD = auto()


@group(Group.ANIMATED)
class AnimatedComponent:  # Component (mixin class)
    """`AnimatedComponent` mixin class for node.

    It provides animation controls,
    and enables the node to manage and play animations defined in `AnimationSet`.

    Attributes:
        `animations`: `AnimationSet` - Collection of named animations.
        `current_animation`: `Animation | None` - Currently active (or paused) animation.
        `repeat`: `bool` - Whether the animation should loop.
        `is_playing`: `bool` - Whether the animation is currently playing.

    Methods:
        `add_animation`
        `play`
        `play_backwards`
        `progress_animation`
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        if (class_animations := getattr(instance, "animations", None)) is not None:
            instance.animations = deepcopy(class_animations)
        else:
            instance.animations = AnimationSet()
        return instance

    animations: AnimationSet
    current_animation: Animation | None = None
    repeat: bool = False
    is_playing: bool = False
    _frame_index: int = 0
    _playback_direction: PlaybackDirection = PlaybackDirection.FORAWRD
    _is_on_last_frame: bool = False

    def with_animations(self, /, **animations: Animation) -> Self:
        """Chained method to add multiple animations.

        Args:
            animations (**Animation): Named animations as keyword arguments.

        Returns:
            Self: Same node instance.
        """
        for animation_name, animation in animations.items():
            setattr(self.animations, animation_name, animation)
        return self

    def with_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> Self:
        """Chained method to add a single animation.

        Args:
            animation_name (str): Name of the animation.
            animation (Animation): Animation instance to add.

        Returns:
            Self: Same node instance.
        """
        self.add_animation(animation_name, animation)
        return self

    def with_repeat(self, state: bool = True, /) -> Self:
        """Chained method to set the repeat state of the animation.

        Args:
            state (bool): Whether the animation should repeat. Defaults to `True`.
        """
        self.repeat = state
        return self

    def add_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> None:
        """Add an animation to the node.

        Args:
            animation_name (str): Name of the animation.
            animation (Animation): Animation instance to add.
        """
        setattr(self.animations, animation_name, animation)

    def play(self, animation_name: str, /) -> None:
        """Play an animation by its name.

        Args:
            animation_name (str): Name of the animation to play.

        Raises:
            ValueError: If the animation with the given name does not exist.
        """
        if not hasattr(self.animations, animation_name):
            raise ValueError(f"Animation not found: '{animation_name}'")
        self.current_animation = getattr(self.animations, animation_name)
        self.is_playing = True
        self._playback_direction = PlaybackDirection.FORAWRD
        self._is_on_last_frame = False
        self._frame_index = 0
        # The actual logic of playing the animation
        # is handled in `.progress_animation`

    def play_backwards(self, animation_name: str, /) -> None:
        """Play an animation in reverse by its name.

        Args:
            animation_name (str): Name of the animation to play in reverse.

        Raises:
            ValueError: If the animation with the given name does not exist.
        """
        if not hasattr(self.animations, animation_name):
            raise ValueError(f"Animation not found: '{animation_name}'")
        self.current_animation = getattr(self.animations, animation_name)
        assert isinstance(self.current_animation, Animation)
        self.is_playing = True
        self._playback_direction = PlaybackDirection.BACKWARD
        self._is_on_last_frame = False
        self._frame_index = len(self.current_animation.frames) - 1
        # The actual logic of playing the animation
        # is handled in `.progress_animation`

    def progress_animation(self) -> None:
        """Progress `1` frame of current animation.

        Called by a frame task, which is found in `Scene.frame_tasks`.
        """
        if self.current_animation is None:
            self.is_playing = False
            return

        # Change texture to the current frame
        self.texture = self.current_animation.frames[self._frame_index]
        frame_count = len(self.current_animation.frames)
        index_change = 1 if self._playback_direction is PlaybackDirection.FORAWRD else -1

        if self.is_playing and self.repeat and self._is_on_last_frame:
            self._is_on_last_frame = False
            first_index = (
                0
                if self._playback_direction is PlaybackDirection.FORAWRD
                else frame_count - 1
            )
            self._frame_index = first_index
            return
        # Progress frame index
        self._frame_index = clamp(
            self._frame_index + index_change,
            0,
            frame_count - 1,
        )
        # Check if hit end of animation, last frame index
        last_index = (
            frame_count - 1
            if self._playback_direction is PlaybackDirection.FORAWRD
            else 0
        )
        # State variable to ensure last frame is shown when `.repeat` is `True`
        if self._frame_index == last_index:
            if self.repeat:
                self._is_on_last_frame = True
            else:
                self.is_playing = False



from __future__ import annotations

import sys
from enum import unique

import charz_core


# NOTE: Variants of the enum produces the same hash as if it was using normal `str`
if sys.version_info >= (3, 11):
    from enum import StrEnum, auto

    @unique
    class Group(StrEnum):
        """Enum containing node groups used in `charz`."""

        NODE = charz_core.Group.NODE
        TEXTURE = auto()
        ANIMATED = auto()
        COLLIDER = auto()
        MOVEMENT = auto()

else:
    from enum import Enum

    @unique
    class Group(str, Enum):
        """Enum containing node groups used in `charz`."""

        NODE = charz_core.Group.NODE.value
        TEXTURE = "texture"
        ANIMATED = "animated"
        COLLIDER = "collider"
        MOVEMENT = "movement"
